from __future__ import annotations

from importlib import resources
from typing import TYPE_CHECKING, no_type_check

import numpy as np

from napari.layers import Points, Shapes
from napari.layers.utils.string_encoding import ConstantStringEncoding

if TYPE_CHECKING:
    from vispy.visuals.text import Text


def update_text(
    *,
    node: Text,
    layer: Points | Shapes,
):
    """Update the vispy text node with a layer's text parameters.

    Parameters
    ----------
    node : vispy.scene.visuals.Text
        The text node to be updated.
    layer : Union[Points, Shapes]
        A layer with text.
    """

    ndisplay = layer._slice_input.ndisplay

    # Vispy always needs non-empty values and coordinates, so if a layer
    # effectively has no visible text then return single dummy data.
    # This also acts as a minor optimization.
    if _has_visible_text(layer):
        text_values = layer._view_text
        colors = layer._view_text_color
        coords, anchor_x, anchor_y = layer._view_text_coords
    else:
        text_values = np.array([''])
        colors = np.zeros((4,), np.float32)
        coords = np.zeros((1, ndisplay))
        anchor_x = 'center'
        anchor_y = 'center'

    # Vispy wants (x, y) positions instead of (row, column) coordinates.
    if ndisplay == 2:
        positions = np.flip(coords, axis=1)
    elif ndisplay == 3:
        raw_positions = np.flip(coords, axis=1)
        n_positions, position_dims = raw_positions.shape

        if position_dims < 3:
            padded_positions = np.zeros((n_positions, 3))
            padded_positions[:, 0:2] = raw_positions
            positions = padded_positions
        else:
            positions = raw_positions

    node.text = text_values
    node.pos = positions
    node.anchors = (anchor_x, anchor_y)

    text_manager = layer.text
    node.rotation = text_manager.rotation
    node.color = colors

    node.font_size = text_manager._get_scaled_size(layer.scale_factor)


def _has_visible_text(layer: Points | Shapes) -> bool:
    text = layer.text
    if not text.visible:
        return False
    if (
        isinstance(text.string, ConstantStringEncoding)
        and text.string.constant == ''
    ):
        return False
    return len(layer._indices_view) != 0


def get_text_width_height(text: Text) -> tuple[float, float]:
    """Get the width and height of a vispy text visual in screen pixels.

    If display scaling is not 1 (e.g. hidpi), this is already accounted for
    by vispy.
    """
    if isinstance(text.text, str):
        strings = [text.text]
    elif isinstance(text.text, list):
        strings = text.text
    else:
        raise TypeError('Text should either be a string or a list of strings')

    top_left_corners = []
    bottom_right_corners = []
    for string in strings:
        if string == '':
            continue

        # this is a private vispy function that calculates the Vertex Buffer Object
        # for the text to be sent to the shaders. This normallly only happens on
        # the fly when rendering, but we need it here because it allows us to
        # calculate the bounding box of the text as it would be when rendered
        buffer = _text_to_vbo(
            string, text._font, *text._anchors, text._font._lowres_size
        )

        pos = buffer['a_position']
        top_left_corners.append(pos.min(axis=0))
        bottom_right_corners.append(pos.max(axis=0))

    top_left = np.min(top_left_corners, axis=0) if top_left_corners else (0, 0)
    bottom_right = (
        np.max(bottom_right_corners, axis=0)
        if bottom_right_corners
        else (0, 0)
    )

    # these magic numbers (1.2 and 1.3) are from trial and error
    return (bottom_right[0] - top_left[0]) * text.font_size, (
        bottom_right[1] - top_left[1]
    ) * text.font_size


# vendored from vispy/visuals/text/text.py, but removing all lines referring to context flushing,
# canvas and viewport which cause issues. We only need to calculate sizes anyways.
@no_type_check
def _text_to_vbo(text, font, anchor_x, anchor_y, lowres_size):
    """Convert text characters to VBO"""
    text_vtype = np.dtype(
        [('a_position', np.float32, 2), ('a_texcoord', np.float32, 2)]
    )
    vertices = np.zeros(len(text) * 4, dtype=text_vtype)
    prev = None
    width = height = ascender = descender = 0
    ratio, slop = 1.0 / font.ratio, font.slop
    x_off = -slop
    # Need to store the original viewport, because the font[char] will
    # trigger SDF rendering, which changes our viewport
    # todo: get rid of call to glGetParameter!

    # Also analyse chars with large ascender and descender, otherwise the
    # vertical alignment can be very inconsistent
    for char in 'hy':
        glyph = font[char]
        y0 = glyph['offset'][1] * ratio + slop
        y1 = y0 - glyph['size'][1]
        ascender = max(ascender, y0 - slop)
        descender = min(descender, y1 + slop)
        height = max(height, glyph['size'][1] - 2 * slop)

    # Get/set the fonts whitespace length and line height (size of this ok?)
    glyph = font[' ']
    spacewidth = glyph['advance'] * ratio
    lineheight = height * 1.5

    # Added escape sequences characters: {unicode:offset,...}
    #   ord('\a') = 7
    #   ord('\b') = 8
    #   ord('\f') = 12
    #   ord('\n') = 10  => linebreak
    #   ord('\r') = 13
    #   ord('\t') = 9   => tab, set equal 4 whitespaces?
    #   ord('\v') = 11  => vertical tab, set equal 4 linebreaks?
    # If text coordinate offset > 0 -> it applies to x-direction
    # If text coordinate offset < 0 -> it applies to y-direction
    esc_seq = {7: 0, 8: 0, 9: -4, 10: 1, 11: 4, 12: 0, 13: 0}

    # Keep track of y_offset to set lines at right position
    y_offset = 0

    # When a line break occur, record the vertices index value
    vi_marker = 0
    ii_offset = 0  # Offset since certain characters won't be drawn

    # The running tracker of characters vertex index
    vi = 0

    for ii, char in enumerate(text):
        if ord(char) in esc_seq:
            if esc_seq[ord(char)] < 0:
                # Add offset in x-direction
                x_off += abs(esc_seq[ord(char)]) * spacewidth
                width += abs(esc_seq[ord(char)]) * spacewidth
            elif esc_seq[ord(char)] > 0:
                # Add offset in y-direction and reset things in x-direction
                dx = dy = 0
                if anchor_x == 'right':
                    dx = -width
                elif anchor_x == 'center':
                    dx = -width / 2.0
                vertices['a_position'][vi_marker : vi + 4] += (dx, dy)
                vi_marker = vi + 4
                ii_offset -= 1
                # Reset variables that affects x-direction positioning
                x_off = -slop
                width = 0
                # Add offset in y-direction
                y_offset += esc_seq[ord(char)] * lineheight
        else:
            # For ordinary characters, normal procedure
            glyph = font[char]
            kerning = glyph['kerning'].get(prev, 0.0) * ratio
            x0 = x_off + glyph['offset'][0] * ratio + kerning
            y0 = glyph['offset'][1] * ratio + slop - y_offset
            x1 = x0 + glyph['size'][0]
            y1 = y0 - glyph['size'][1]
            u0, v0, u1, v1 = glyph['texcoords']
            position = [[x0, y0], [x0, y1], [x1, y1], [x1, y0]]
            texcoords = [[u0, v0], [u0, v1], [u1, v1], [u1, v0]]
            vi = (ii + ii_offset) * 4
            vertices['a_position'][vi : vi + 4] = position
            vertices['a_texcoord'][vi : vi + 4] = texcoords
            x_move = glyph['advance'] * ratio + kerning
            x_off += x_move
            ascender = max(ascender, y0 - slop)
            descender = min(descender, y1 + slop)
            width += x_move
            height = max(height, glyph['size'][1] - 2 * slop)
            prev = char

    dx = dy = 0
    if anchor_y == 'top':
        dy = -descender
    elif anchor_y in ('center', 'middle'):
        dy = (-descender - ascender) / 2
    elif anchor_y == 'bottom':
        dy = -ascender
    if anchor_x == 'right':
        dx = -width
    elif anchor_x == 'center':
        dx = -width / 2.0

    # If any linebreaks occured in text, we only want to translate characters
    # in the last line in text (those after the vi_marker)
    vertices['a_position'][0:vi_marker] += (0, dy)
    vertices['a_position'][vi_marker:] += (dx, dy)
    vertices['a_position'] /= lowres_size

    return vertices


def register_napari_fonts() -> None:
    from vispy.util.fonts import register_vispy_font

    path = resources.files('napari') / 'resources' / 'fonts' / 'AlataPlus'
    register_vispy_font(path, 'AlataPlus', False, False)



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.layer_controls.widgets import (
    QtOutSliceCheckBoxControl,
    QtProjectionModeControl,
)
from napari._qt.layer_controls.widgets._vectors import (
    QtEdgeColorFeatureControl,
    QtLengthSpinBoxControl,
    QtVectorStyleComboBoxControl,
    QtWidthSpinBoxControl,
)
from napari.layers.base._base_constants import Mode

if TYPE_CHECKING:
    import napari.layers


class QtVectorsControls(QtLayerControls):
    """Qt view and controls for the napari Vectors layer.

    Parameters
    ----------
    layer : napari.layers.Vectors
        An instance of a napari Vectors layer.

    Attributes
    ----------
    _edge_color_feature_control : napari._qt.layer_controls.widgets._vectors.QtEdgeColorFeatureControl
        Widget that wraps the widgets used to select vectors edge color mode, feature and color.
    _length_spinbox_control : napari._qt.layer_controls.widgets._vectors.QtLengthSpinBoxControl
        Widget that wraps a spinbox widget controlling length of vectors.
    _out_slice_checkbox_control : napari._qt.layer_controls.widgets.QtOutSliceCheckBoxControl
        Widget that wraps a checkbox to indicate whether to render out of slice.
    _projection_mode_control : napari._qt.layer_controls.widgets.QtProjectionModeControl
        Widget that wraps dropdown menu to select the projection mode for the layer.
    _vector_style_combobox_control : napari._qt.layer_controls.widgets._vectors.QtVectorStyleComboBoxControl
        Widget that wraps a dropdown widget to select vector_style for the vectors.
    _width_spinbox_control : napari._qt.layer_controls.widgets._vectors.QtWidthSpinBoxControl
        Widget that wraps a spinbox controlling edge line width of vectors.
    """

    layer: 'napari.layers.Vectors'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_tracks_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_tracks_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)
        # Setup widgets controls
        self._width_spinbox_control = QtWidthSpinBoxControl(self, layer)
        self._add_widget_controls(self._width_spinbox_control)
        self._length_spinbox_control = QtLengthSpinBoxControl(self, layer)
        self._add_widget_controls(self._length_spinbox_control)
        self._projection_mode_control = QtProjectionModeControl(self, layer)
        self._add_widget_controls(self._projection_mode_control)
        self._vector_style_combobox_control = QtVectorStyleComboBoxControl(
            self, layer
        )
        self._add_widget_controls(self._vector_style_combobox_control)
        self._edge_color_feature_control = QtEdgeColorFeatureControl(
            self, layer
        )
        self._add_widget_controls(self._edge_color_feature_control)
        self._out_slice_checkbox_control = QtOutSliceCheckBoxControl(
            self, layer
        )
        self._add_widget_controls(self._out_slice_checkbox_control)



from napari._qt.layer_controls.widgets._image.qt_depiction_control import (
    QtDepictionControl,
)
from napari._qt.layer_controls.widgets._image.qt_interpolation_combobox import (
    QtInterpolationComboBoxControl,
)
from napari._qt.layer_controls.widgets._image.qt_render_control import (
    QtImageRenderControl,
)

__all__ = [
    'QtDepictionControl',
    'QtImageRenderControl',
    'QtInterpolationComboBoxControl',
]



import datetime
import uuid
from enum import auto
from typing import Optional

from napari.utils.misc import Callable, StringEnum


class Status(StringEnum):
    PENDING = auto()
    BUSY = auto()
    COMPLETED = auto()
    CANCELLED = auto()
    FAILED = auto()
    START_FAILED = auto()


class TaskStatusItem:
    def __init__(
        self,
        provider: str,
        status: Status,
        description: str,
        cancel_callback: Optional[Callable] = None,
    ) -> None:
        self.id: uuid.UUID = uuid.uuid4()
        self._provider = provider
        self._timestamp = [self._timestap()]
        self._status = [status]
        self._description = [description]
        self._cancel_callback = cancel_callback

    def _timestap(self) -> str:
        return datetime.datetime.now().isoformat()

    def __str__(self) -> str:
        return f'TaskStatusItem: ({self._provider}, {self.id}, {self._timestamp[-1]}, {self._status[-1]}, {self._description[-1]})'

    def update(self, status: Status, description: str) -> None:
        self._timestamp.append(self._timestap())
        self._status.append(status)
        self._description.append(description)

    def cancel(self) -> bool:
        self.update(Status.CANCELLED, '')
        if self._cancel_callback is not None:
            return self._cancel_callback()
        return False

    def state(self) -> tuple[str, str, Status, str]:
        return (
            self._provider,
            self._timestamp[-1],
            self._status[-1],
            self._description[-1],
        )


class TaskStatusManager:
    """
    A task status manager, to store status of long running processes/tasks.

    Only one instance is in general available through napari.

    napari methods and plugins can use it to register and update
    long running tasks.
    """

    _tasks: dict[uuid.UUID, TaskStatusItem]

    def __init__(self) -> None:
        # Note: we are using a dict here that may not be thread-safe; however
        # given the that the values from it are added/updated using an UUID
        # collision chances are low and it should be ok as long as operations
        # that require its iteration (`is_busy`, `get_status`, `cancel_all`)
        # are done when no task status additions are scheduled (i.e when
        # closing the application).
        self._tasks: dict[uuid.UUID, TaskStatusItem] = {}

    def register_task_status(
        self,
        provider: str,
        task_status: Status,
        description: str,
        cancel_callback: Optional[Callable] = None,
    ) -> uuid.UUID:
        item = TaskStatusItem(
            provider, task_status, description, cancel_callback
        )
        self._tasks[item.id] = item
        return item.id

    def update_task_status(
        self,
        status_id: uuid.UUID,
        task_status: Status,
        description: str = '',
    ) -> bool:
        if status_id in self._tasks:
            item = self._tasks[status_id]
            item.update(task_status, description)
            return True

        return False

    def is_busy(self) -> bool:
        for _, item in self._tasks.items():
            if item.state()[2] in {Status.PENDING, Status.BUSY}:
                return True
        return False

    def get_status(self) -> list[str]:
        messages = []
        for _, item in self._tasks.items():
            provider, _ts, status, description = item.state()
            if status in {Status.PENDING, Status.BUSY}:
                messages.append(f'{provider}: {description}')

        return messages

    def cancel_all(self) -> None:
        for _, item in self._tasks.items():
            item.cancel()



from __future__ import annotations

import logging
import re
import textwrap
import warnings
from typing import TYPE_CHECKING, Any

import numpy as np
from vispy.scene.node import Node
from vispy.scene.visuals import Polygon
from vispy.util.svg import Document
from vispy.visuals.transforms import STTransform

from napari._app_model import get_app_model
from napari._vispy.visuals.text import Text
from napari.resources import get_icon_path
from napari.settings import get_settings
from napari.utils.action_manager import action_manager
from napari.utils.interactions import Shortcut

if TYPE_CHECKING:
    from napari.utils.color import ColorValue

vispy_logger = logging.getLogger('vispy')


class Welcome(Node):
    def __init__(self) -> None:
        old_level = vispy_logger.level
        vispy_logger.setLevel(logging.ERROR)
        self.logo_coords = (
            Document(get_icon_path('logo_silhouette')).paths[0].vertices[0][0]
        )
        vispy_logger.setLevel(old_level)
        self.logo_coords = self.logo_coords[
            :, :2
        ]  # drop z: causes issues with polygon agg mode
        # center
        self.logo_coords -= (
            np.max(self.logo_coords, axis=0) + np.min(self.logo_coords, axis=0)
        ) / 2
        # make it smaller
        self.logo_coords /= 4
        # move it up
        self.logo_coords[:, 1] -= 130  # magic number shifting up logo
        super().__init__()

        self.logo = Polygon(
            self.logo_coords, border_method='agg', border_width=2, parent=self
        )
        self.header = Text(
            text='',
            pos=[0, 0],
            anchor_x='center',
            anchor_y='bottom',
            method='gpu',
            parent=self,
        )
        self.shortcut_keybindings = Text(
            text='',
            line_height=1.15,
            pos=[-80, 60],
            anchor_x='right',
            anchor_y='bottom',
            method='gpu',
            parent=self,
        )
        self.shortcut_descriptions = Text(
            text='',
            line_height=1.15,
            pos=[-60, 60],
            anchor_x='left',
            anchor_y='bottom',
            method='gpu',
            parent=self,
        )
        self.tip = Text(
            text='',
            line_height=1.15,
            pos=[0, 160],
            anchor_x='center',
            anchor_y='bottom',
            method='gpu',
            parent=self,
        )

        self.transform = STTransform()

    def set_color(self, color: ColorValue) -> None:
        self.logo.color = color
        self.logo.border_color = color
        self.header.color = color
        self.shortcut_keybindings.color = color
        self.shortcut_descriptions.color = color
        self.tip.color = color

    def set_version(self, version: str) -> None:
        self.header.text = (
            f'napari {version}\n\n'
            'Drag file(s) here to open, or use the shortcuts below:'
        )

    def set_shortcuts(self, commands: tuple[str, ...]) -> None:
        shortcuts = {}
        for command_id in commands:
            shortcut, command = self._command_shortcut_and_description(
                command_id
            )
            if shortcut is not None and command is not None:
                shortcuts[shortcut] = command

        # TODO: use template strings in the future
        self.shortcut_keybindings.text = '\n'.join(shortcuts.keys())
        self.shortcut_descriptions.text = '\n'.join(shortcuts.values())

    def set_tip(self, tip: str) -> None:
        # TODO: this should use template strings in the future
        for match in re.finditer(r'{(.*?)}', tip):
            command_id = match.group(1)
            shortcut, _ = self._command_shortcut_and_description(command_id)
            # this can be none at launch (not yet initialized), will be updated after
            if shortcut is None:
                # maybe it was just a direct keybinding given
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    shortcut = Shortcut(command_id).platform
            if shortcut:
                tip = re.sub(match.group(), str(shortcut), tip)

        # wrap tip so it's not clipped
        self.tip.text = 'Did you know?\n' + '\n'.join(
            textwrap.wrap(tip, break_on_hyphens=False)
        )

    @staticmethod
    def _command_shortcut_and_description(
        command_id: str,
    ) -> tuple[str | None, str | None]:
        app = get_app_model()
        all_shortcuts = get_settings().shortcuts.shortcuts
        keybinding = app.keybindings.get_keybinding(command_id)

        shortcut = command = None
        if keybinding is not None:
            shortcut = Shortcut(keybinding.keybinding).platform
            command = app.commands[command_id].title
        else:
            # might be an action_manager action
            keybinding = all_shortcuts.get(command_id, [None])[0]
            if keybinding is not None:
                shortcut = Shortcut(keybinding).platform
                command = action_manager._actions[command_id].description
            else:
                shortcut = command = None

        return shortcut, command

    def set_scale_and_position(self, x: float, y: float) -> None:
        self.transform.translate = (x / 2, y / 2, 0, 0)
        scale = min(x, y) * 0.002  # magic number
        self.transform.scale = (scale, scale, 0, 0)

        for text in (
            self.header,
            self.shortcut_keybindings,
            self.shortcut_descriptions,
            self.tip,
        ):
            text.font_size = max(scale * 10, 10)

    def set_gl_state(self, *args: Any, **kwargs: Any) -> None:
        for node in self.children:
            node.set_gl_state(*args, **kwargs)



"""
Add shapes with text
====================

Display one shapes layer ontop of one image layer using the ``add_shapes`` and
``add_image`` APIs. When the window is closed it will print the coordinates of
your shapes.

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data

import napari

# add the image
viewer = napari.Viewer()
layer = viewer.add_image(data.camera(), name='photographer')

# create a list of polygons
polygons = [
    np.array([[225, 146], [283, 146], [283, 211], [225, 211]]),
    np.array([[67, 182], [167, 182], [167, 268], [67, 268]]),
    np.array([[111, 336], [220, 336], [220, 240], [111, 240]]),
]

# create features
features = {
    'likelihood': [21.23423, 51.2315, 100],
    'class': ['hand', 'face', 'camera'],
}
edge_color_cycle = ['blue', 'magenta', 'green']

text = {
    'string': '{class}: {likelihood:0.1f}%',
    'anchor': 'upper_left',
    'translation': [-5, 0],
    'size': 8,
    'color': 'green',
}

# add polygons
shapes_layer = viewer.add_shapes(
    polygons,
    features=features,
    shape_type='polygon',
    edge_width=3,
    edge_color='class',
    edge_color_cycle=edge_color_cycle,
    face_color='transparent',
    text=text,
    name='shapes',
)

# change some attributes of the layer
shapes_layer.opacity = 1

# To save layers to svg:
# viewer.layers.save('viewer.svg', plugin='svg')

if __name__ == '__main__':
    napari.run()



import itertools
from typing import Any

import numpy as np
import numpy.typing as npt
from vispy.scene.visuals import Compound, Image
from vispy.visuals import BaseVisual
from vispy.visuals.transforms.linear import STTransform

# attributes that are passed from the tiled node to the child image nodes.
PASS_THROUGH_ATTRIBUTES = {
    'cmap',
    'clim',
    'opacity',
    'gamma',
    'interpolation',
    'custom_kernel',
}


class TiledImageNode(Compound):
    """Custom Vispy scenegraph Node to display large images.

    Vispy 2D rendering works by drawing images as 2D OpenGL textures.
    OpenGL has a texture size limit (driver and hardware dependent), which
    means that some images are too large to be displayed by a single
    texture. This class automatically splits those images into tiles not
    exceeding the texture size, and displays each tile as a texture, offset
    by the appropriate amount in visual space.

    Attributes such as colormap and contrast limits are passed through to
    the child nodes using setattr.

    Attributes
    ----------
    data : np.ndarray
        The data to be displayed.
    texture_format : str
        The texture format (uint8, uint16, float32).
    tile_size : int
        The maximum tile size, used to split the image.
    adopted_children : list[Image]
        The child Image nodes containing the component tiles of the image.
        Note: we use the term "adopted children" because "children" is a
        protected attribute of the Node class, and contains other VisualNodes.
    offsets : list[tuple[int, int]]
        The x/y offset of each tile (in the same order as `adopted_children`).
    """

    def __init__(
        self,
        data: np.ndarray,
        tile_size: int,
        texture_format: str | None = None,
    ) -> None:
        self.unfreeze()
        self.texture_format = texture_format
        self.adopted_children: list[Image] = []
        self.offsets: list[tuple[int, int]] = []
        self.tile_size = tile_size
        self.data: npt.ArrayLike | None = None
        super().__init__([])
        self.set_data(data)

    def add_subvisual(self, visual: BaseVisual) -> None:
        """Override for CompoundVisual to not update subvisual transforms.

        CompoundVisual normally sets the transforms for all subvisuals to be
        the same as the parent as they are added to the CompoundVisual. For the
        tiled image layer, that is totally not what we want! So we make sure to
        just update the parent.

        See:
        https://github.com/napari/napari/pull/8395#issuecomment-3776882524
        """
        self._subvisuals.append(visual)

    def _transform_changed(self, event: Any = None) -> None:
        """Override for CompoundVisual to not update subvisual transforms.

        CompoundVisual normally sets the transforms for all subvisuals to be
        the same as the parent. For the tiled image layer, that is totally not
        what we want! So we make sure to just update the parent.

        See:
        https://github.com/napari/napari/pull/8395#issuecomment-3776882524
        """
        BaseVisual._transform_changed(self)

    def handle_axis_change(self) -> None:
        """Handle changes in axis order by updating tile transforms."""

        for ch, offset in zip(
            self.adopted_children, self.offsets, strict=True
        ):
            ch.transform = STTransform(translate=offset + (0,))

    def set_data(self, data: np.ndarray) -> None:
        tiles = make_tiles(data, self.tile_size)
        self.offsets = [of for of, _ in tiles]
        self.data = data
        # if the correct number of tiles already exist, just update their data
        # otherwise, delete all existing tiles and create new ones
        if len(self.adopted_children) == len(tiles):
            for ch, (_, dat) in zip(self.adopted_children, tiles, strict=True):
                ch.set_data(dat)
        else:
            for child in self.adopted_children:
                child.parent = None
            self._subvisuals: list[BaseVisual] = []
            self.adopted_children = [
                Image(
                    data=dat,
                    parent=self,
                    texture_format=self.texture_format,
                )
                for _, dat in tiles
            ]
            for ch, offset in zip(
                self.adopted_children, self.offsets, strict=True
            ):
                ch.transform = STTransform(translate=offset + (0,))
                self.add_subvisual(ch)

    def set_gl_state(self, *args: Any, **kwargs: Any) -> None:
        for child in self.adopted_children:
            child.set_gl_state(*args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        if name in PASS_THROUGH_ATTRIBUTES | {'events'}:
            if len(self.adopted_children) > 0:
                return getattr(self.adopted_children[0], name)
            return None
        return self.__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in PASS_THROUGH_ATTRIBUTES:
            for child in self.adopted_children:
                setattr(child, name, value)
        else:
            super().__setattr__(name, value)


def make_tiles(
    image: np.ndarray, tile_size: int
) -> list[tuple[tuple[int, int], np.ndarray]]:
    """Split a large image into a list of tiles and offsets.

    Supports input shapes (H, W) for grayscale or (H, W, 3) for RGB.

    Parameters
    ----------
    image : np.ndarray, shape (H, W[, 3])
        The input image.
    tile_size : int
        The maximum size of any tile along any axis.

    Returns
    -------
    tile_list : list[tuple[tuple[int, int], np.ndarray]]
        List of x, y offsets and corresponding array tiles.
    """
    h, w, *_ = image.shape

    tile_list = [
        # note: vispy space is transposed (y, x -> x, y) compared to NumPy
        # indexing space; we want the vispy offsets so we swap them here.
        ((x, y), image[y : y + tile_size, x : x + tile_size])
        for y, x in itertools.product(
            range(0, h, tile_size), range(0, w, tile_size)
        )
    ]

    return tile_list



"""
Add shapes with features
========================

Display one shapes layer ontop of one image layer using the ``add_shapes`` and
``add_image`` APIs. When the window is closed it will print the coordinates of
your shapes.

.. tags:: visualization-basic, features-table
"""

import numpy as np
from skimage import data

import napari

# add the image
viewer = napari.Viewer()
layer = viewer.add_image(data.camera(), name='photographer')

# create a list of polygons
polygons = [
    np.array([[11, 13], [111, 113], [22, 246]]),
    np.array(
        [
            [505, 60],
            [402, 71],
            [383, 42],
            [251, 95],
            [212, 59],
            [131, 137],
            [126, 187],
            [191, 204],
            [171, 248],
            [211, 260],
            [273, 243],
            [264, 225],
            [430, 173],
            [512, 160],
        ]
    ),
    np.array(
        [
            [310, 382],
            [229, 381],
            [209, 401],
            [221, 411],
            [258, 411],
            [300, 412],
            [306, 435],
            [268, 434],
            [265, 454],
            [298, 461],
            [307, 461],
            [307, 507],
            [349, 510],
            [352, 369],
            [330, 366],
            [330, 366],
        ]
    ),
]

# create features
features = {
    'likelihood': [0.2, 0.5, 1],
    'class': ['sky', 'person', 'building'],
}
face_color_cycle = ['blue', 'magenta', 'green']

# add polygons
layer = viewer.add_shapes(
    polygons,
    features=features,
    shape_type='polygon',
    edge_width=1,
    edge_color='likelihood',
    edge_colormap='gray',
    face_color='class',
    face_color_cycle=face_color_cycle,
    name='shapes',
)

# change some attributes of the layer
layer.selected_data = set(range(layer.nshapes))
layer.current_edge_width = 5
layer.selected_data = set()

# To save layers to svg:
# viewer.layers.save('viewer.svg', plugin='svg')

if __name__ == '__main__':
    napari.run()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from napari._vispy.overlays.base import LayerOverlayMixin, VispyCanvasOverlay
from napari._vispy.visuals.colormap import Colormap
from napari.settings import get_settings
from napari.utils.colormaps.colormap_utils import (
    _coerce_contrast_limits,
    _napari_cmap_to_vispy,
)
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.theme import get_theme

if TYPE_CHECKING:
    from vispy.scene import Node

    from napari.components.overlays import Overlay
    from napari.layers import Image, Surface


class VispyColorBarOverlay(LayerOverlayMixin, VispyCanvasOverlay):
    def __init__(
        self,
        *,
        layer: Image | Surface,
        overlay: Overlay,
        parent: Node | None = None,
    ) -> None:
        super().__init__(
            node=Colormap(), layer=layer, overlay=overlay, parent=parent
        )
        self.layer: Image | Surface
        self.x_size = 50
        self.y_size = 250
        self.x_offset = 7
        self.y_offset = 7

        self.layer.events.contrast_limits.connect(self._on_data_change)
        self.layer.events.colormap.connect(self._on_colormap_change)
        self.layer.events.gamma.connect(self._on_gamma_change)
        self.overlay.events.size.connect(self._on_size_change)
        self.overlay.events.tick_length.connect(self._on_ticks_change)
        self.overlay.events.font_size.connect(self._on_ticks_change)
        self.overlay.events.color.connect(self._on_ticks_change)

        get_settings().appearance.events.theme.connect(self._on_data_change)

        self.reset()

    def _on_data_change(self) -> None:
        self.node.set_data_and_clim(
            clim=_coerce_contrast_limits(
                self.layer.contrast_limits
            ).contrast_limits,
            dtype=self.layer.dtype,
        )
        self._on_colormap_change()
        self._on_gamma_change()
        self._on_ticks_change()

    def _on_colormap_change(self) -> None:
        self.node.set_cmap(_napari_cmap_to_vispy(self.layer.colormap))

    def _on_gamma_change(self) -> None:
        self.node.set_gamma(self.layer.gamma)

    def _on_size_change(self) -> None:
        self.node.set_size(self.overlay.size)
        self._on_ticks_change()

    def _on_ticks_change(self) -> None:
        # set color to the negative of theme background.
        # the reason for using the `as_hex` here is to avoid
        # `UserWarning` which is emitted when RGB values are above 1
        color = self.overlay.color
        if color is None:
            if (
                self.node.parent is not None
                and self.node.parent.canvas.bgcolor
            ):
                background_color = self.node.parent.canvas.bgcolor.rgba
            else:
                background_color = get_theme(
                    get_settings().appearance.theme
                ).canvas.as_hex()
                background_color = transform_color(background_color)[0]
            color = np.subtract(1, background_color)
            color[-1] = background_color[-1]

        text_width, text_height = self.node.set_ticks_and_get_text_size(
            tick_length=self.overlay.tick_length,
            font_size=self.overlay.font_size,
            clim=_coerce_contrast_limits(
                self.layer.contrast_limits
            ).contrast_limits,
            color=color,
        )

        # Calculate proper layout with explicit spacing constants
        self.x_size = (
            self.overlay.size[0]  # Colorbar width
            + self.overlay.tick_length  # Tick marks length
            + text_width  # Text width with margins
        )
        self.y_size = text_height

        self._on_position_change()

    def reset(self) -> None:
        super().reset()
        self._on_data_change()
        self._on_size_change()



"""
Concentric spheres
==================

Display concentric spheres in 3D.

.. tags:: visualization-nD
"""

import numpy as np
from skimage import morphology

import napari

b0 = morphology.ball(5)

b1 = morphology.ball(10)

b0p = np.pad(b0, 5)

viewer = napari.Viewer(ndisplay=3)

# viewer.add_labels(b0)
viewer.add_labels(b0p)
viewer.add_labels(b1 * 2)

viewer.add_points([[10, 10, 10]], size=1)

if __name__ == '__main__':
    napari.run()



"""
Surface normals wireframe
=========================

Display a 3D mesh with normals and wireframe

.. tags:: experimental
"""

from vispy.io import load_data_file, read_mesh

import napari

vert, faces, _, _ = read_mesh(load_data_file('orig/triceratops.obj.gz'))

# put the mesh right side up, scale it up (napari#3477) and fix faces handedness
vert *= -100
faces = faces[:, ::-1]

viewer = napari.Viewer(ndisplay=3)
surface = viewer.add_surface(data=(vert, faces))

# enable normals and wireframe
surface.normals.face.visible = True
surface.normals.vertex.visible = True
surface.wireframe.visible = True

if __name__ == '__main__':
    napari.run()



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Polygon

from napari.layers.shapes._accelerated_triangulate_dispatch import (
    generate_2D_edge_meshes,
)

fig, axes = plt.subplots(2, 3)
# fig.set_figwidth(15)
# fig.set_figheight(10)
colors = iter(['red', 'green', 'blue', 'yellow'])
itaxes = iter(axes.flatten())
sup = axes.flatten()[4]
for closed in [False, True]:
    for beveled in [False, True]:
        ax = next(itaxes)
        c = next(colors)
        centers, offsets, triangles = generate_2D_edge_meshes(
            np.array([[0, 3], [1, 0], [2, 3], [5, 0], [2.5, 5]]),
            closed=closed,
            limit=3,
            bevel=beveled,
        )
        points = centers + 0.3 * offsets
        for t in triangles:
            trp = points[t]
            ax.add_patch(Polygon(trp, ec='#000000', fc=c, alpha=0.2))
            sup.add_patch(Polygon(trp, ec='#000000', fc=c, alpha=0.1))
        ax.scatter(*(points).T)
        ax.scatter(*(centers).T)
        ax.set_aspect('equal')
        ax.set_title(f' {closed=}, {beveled=}')
        ax.set_xlim(-1, 6)
        ax.set_ylim(-1, 6)
        sup.set_xlim(-1, 6)
        sup.set_ylim(-1, 6)

if __name__ == '__main__':
    plt.show()



import logging
import os
import pathlib
import shutil
import subprocess
from argparse import ArgumentParser

logging.basicConfig(
    format='%(levelname)s : %(asctime)s : %(message)s',
    level=logging.INFO,
)

parser = ArgumentParser(
    description='Run napari with one of the perfmon configurations.'
)
parser.add_argument(
    'config',
    help='The name of the sub-directory that contains the perfmon configuration file (e.g. slicing).',
)
parser.add_argument(
    'example_script', help='The example script that should run napari.'
)
parser.add_argument(
    '--output',
    default='latest',
    help='The name to add to the output traces file.',
)
args = parser.parse_args()

logging.info(
    'Running run.py with the following arguments.\n{args_}',
    extra={'args_': args},
)

perfmon_dir = pathlib.Path(__file__).parent.resolve(strict=True)
config_dir = perfmon_dir / args.config
config_path = str(config_dir / 'config.json')

env = os.environ.copy()
env['NAPARI_PERFMON'] = config_path

subprocess.check_call(
    ['python', args.example_script],
    env=env,
)

original_output_path = str(config_dir / 'traces-latest.json')
desired_output_path = str(config_dir / (f'traces-{args.output}.json'))
if desired_output_path != original_output_path:
    shutil.copy(original_output_path, desired_output_path)



# mypy: disable-error-code="attr-defined"
import warnings
from collections.abc import Generator
from copy import deepcopy

import numpy as np
from magicgui import magicgui

import napari
from napari._vispy.mouse_event import NapariMouseEvent
from napari._vispy.overlays.interaction_box import VispySelectionBoxOverlay
from napari._vispy.utils.visual import overlay_to_visual
from napari.components.overlays import SelectionBoxOverlay
from napari.components.overlays.interaction_box import InteractionBoxHandle
from napari.layers import Image
from napari.layers.utils.interaction_box import (
    generate_interaction_box_vertices,
    get_nearby_handle,
)


# define a model for the selection box overlay;
# we subclass from SelectionBoxOverlay to get the
# default behavior of the selection box;
class SelectionBoxNoRotation(SelectionBoxOverlay):
    """Selection box overlay with no rotation handle."""


# we also need to define an equivalent vispy overlay;
# again, we subclass from VispySelectionBoxOverlay
class VispySelectionBoxNoRotation(VispySelectionBoxOverlay):
    """Vispy selection box overlay with no rotation handle."""

    # the _on_bounds_change method is the same as in the
    # original VispySelectionBoxOverlay, but we set
    # rotation to False to not draw the rotation handle
    def _on_bounds_change(self) -> None:
        if self.layer._slice_input.ndisplay == 2:
            top_left, bot_right = self.overlay.bounds
            self.node.set_data(
                # invert axes for vispy
                top_left[::-1],
                bot_right[::-1],
                handles=self.overlay.handles,
                selected=self.overlay.selected_handle,
                # by setting rotation to False,
                # the circle handle will not be drawn
                rotation=False,
            )


# before we can use the new overlay, we have to update
# the overlay_to_visual mapping to include our new overlay;
# this is necessary so that the correct vispy overlay
# is used when the overlay is created
overlay_to_visual[SelectionBoxNoRotation] = VispySelectionBoxNoRotation

viewer = napari.Viewer()

# we add an image layer with random data;
data = np.random.randint(0, 255, size=(1024, 512), dtype=np.uint8)
image = viewer.add_image(
    data,
    name='image',
)

# just for type checking
assert isinstance(image, Image)

# we recover the bounds of the image layer;
# this method will ensure that the overlay is drawn
# correctly in the viewer
# TODO: this half pixel offset should be done on the visual side actually
layer_bounds = ((0, 0), data.shape)
image._overlays['selection_no_rotation'] = SelectionBoxNoRotation(
    bounds=layer_bounds, handles=True
)

# with some adjustments, the selection box overlay
# can be interacted with via mouse events;
# we need to first setup the mouse event handlers
# to allow for the interaction with the overlay;


# this callback will handle the mouse events of
# dragging and dropping the selection box handles;
# it will check if the mouse is in range of one of the
# overlay handles; if it is, we will set the selected handle
# to the handle that is closest to the mouse position;
# then we will set the bounds of the overlay to the
# new position of the mouse;
def resize_selection_box(
    layer: Image, event: 'NapariMouseEvent'
) -> 'Generator[None, None, None]':
    """Resize the selection box based on mouse movement.

    Parameters
    ----------
    layer : DetectorLayer
        The layer to resize the selection box for.
    event : NapariMouseEvent
        The event triggered by mouse movement.

    Yields
    ------
    None
        This is a generator function that handles mouse dragging.
    """
    if len(event.dims_displayed) != 2:
        return

    # Get the selected handle
    selected_handle = layer._overlays['selection_no_rotation'].selected_handle
    if selected_handle is None or selected_handle in [
        InteractionBoxHandle.INSIDE,
        InteractionBoxHandle.ROTATION,
    ]:
        # If no handle is selected or the selected handle
        # is INSIDE or ROTATION, do nothing
        return

    top_left, bot_right = (
        list(x)
        for x in deepcopy(layer._overlays['selection_no_rotation'].bounds)
    )

    layer_bounds = image._display_bounding_box_augmented([0, 1])

    # to prevent the event from being passed down to the
    # pan-zoom event handler, set the event as handled;
    event.handled = True

    yield

    # Main event loop for handling drag events
    while event.type == 'mouse_move':
        mouse_pos = layer.world_to_data(event.position)[event.dims_displayed]
        clipped_y = np.clip(mouse_pos[0], *layer_bounds[0])
        clipped_x = np.clip(mouse_pos[1], *layer_bounds[1])

        # based on the new mouse position, we recalculate the bounds
        # of the overlay; we need to ensure that the new bounds are within
        # the bounds of the image
        match selected_handle:
            case InteractionBoxHandle.TOP_LEFT:
                top_left[0] = clipped_y
                top_left[1] = clipped_x
            case InteractionBoxHandle.TOP_CENTER:
                top_left[0] = clipped_y
            case InteractionBoxHandle.TOP_RIGHT:
                top_left[0] = clipped_y
                bot_right[1] = clipped_x
            case InteractionBoxHandle.CENTER_LEFT:
                top_left[1] = clipped_x
            case InteractionBoxHandle.CENTER_RIGHT:
                bot_right[1] = clipped_x
            case InteractionBoxHandle.BOTTOM_LEFT:
                bot_right[0] = clipped_y
                top_left[1] = clipped_x
            case InteractionBoxHandle.BOTTOM_CENTER:
                bot_right[0] = clipped_y
            case InteractionBoxHandle.BOTTOM_RIGHT:
                bot_right[0] = clipped_y
                bot_right[1] = clipped_x
            case _:
                pass

        # now we update the bounds of the overlay
        # to trigger the visual update;
        layer._overlays['selection_no_rotation'].bounds = deepcopy(
            (tuple(top_left), tuple(bot_right))
        )
        yield


# this callback will hightlight the overlay handles
# when the mouse hovers over them;
def highlight_roi_box_handles(layer: Image, event: NapariMouseEvent) -> None:
    """Highlight the hovered handle of a selection box.

    Parameters
    ----------
    layer : Image
        The layer to highlight the selection box for.
    event : NapariMouseEvent
        The event triggered by mouse movement.
    """
    # the event is not handled by the viewer
    # if the number of displayed dimensions is not 2
    # this is a requirement for the overlay to be displayed
    if len(event.dims_displayed) != 2:
        return

    # we work in data space so we're axis aligned which simplifies calculation
    # same as Layer.world_to_data
    world_to_data = (
        layer._transforms[1:].set_slice(layer._slice_input.displayed).inverse
    )

    # interaction box calculations all happen in vispy coordinates (zyx)
    pos = np.array(world_to_data(event.position))[event.dims_displayed][::-1]

    top_left, bot_right = layer._overlays['selection_no_rotation'].bounds
    handle_coords = generate_interaction_box_vertices(
        top_left[::-1], bot_right[::-1], handles=True
    )
    nearby_handle = get_nearby_handle(pos, handle_coords)

    # if the selected handle is INSIDE or ROTATION, we don't want to
    # highlight the handles, so we return without doing anything
    if nearby_handle in [
        InteractionBoxHandle.INSIDE,
        InteractionBoxHandle.ROTATION,
    ]:
        nearby_handle = None

    # set the selected vertex of the box to the nearby_handle (can also be INSIDE or None)
    layer._overlays['selection_no_rotation'].selected_handle = nearby_handle


# after defining the callbacks, we need to connect them to our layer;
# mouse_move_callbacks is a list of callbacks invoked when the mouse
# hovers over the layer;
# mouse_drag_callbacks is a list of callbacks invoked when the
# mouse is pressed, moved and released;
image.mouse_move_callbacks.append(highlight_roi_box_handles)
image.mouse_drag_callbacks.append(resize_selection_box)


# we use a simple magicgui widget to allow
# the toggling of the selection box overlay
# as demonstration
@magicgui(auto_call=True)
def toggle_overlay(
    viewer: napari.Viewer, toggle_selection_box: bool = False
) -> None:
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        viewer.layers['image']._overlays[
            'selection_no_rotation'
        ].visible = toggle_selection_box


# add the widget to the viewer
viewer.window.add_dock_widget(toggle_overlay)

if __name__ == '__main__':
    napari.run()



"""
bbox annotator
==============

.. tags:: gui
"""

import numpy as np
import pandas as pd
from magicgui.widgets import ComboBox, Container
from skimage import data

import napari

# set up the categorical annotation values and text display properties
box_annotations = ['person', 'sky', 'camera']
text_feature = 'box_label'
features = pd.DataFrame({
    text_feature: pd.Series([], dtype=pd.CategoricalDtype(box_annotations))
})
text_color = 'green'
text_size = 20


# create the GUI for selecting the values
def create_label_menu(shapes_layer, label_feature, labels):
    """Create a label menu widget that can be added to the napari viewer dock

    Parameters
    ----------
    shapes_layer : napari.layers.Shapes
        a napari shapes layer
    label_feature : str
        the name of the shapes feature to use the displayed text
    labels : List[str]
        list of the possible text labels values.

    Returns
    -------
    label_widget : magicgui.widgets.Container
        the container widget with the label combobox
    """
    # Create the label selection menu
    label_menu = ComboBox(label='text label', choices=labels)
    label_widget = Container(widgets=[label_menu])

    def update_label_menu():
        """This is a callback function that updates the label menu when
        the default features of the Shapes layer change
        """
        new_label = str(shapes_layer.feature_defaults[label_feature][0])
        if new_label != label_menu.value:
            label_menu.value = new_label

    shapes_layer.events.feature_defaults.connect(update_label_menu)

    def set_selected_features_to_default():
        """This is a callback that updates the feature values of the currently
        selected shapes. This is a side-effect of the deprecated current_properties
        setter, but does not occur when modifying feature_defaults."""
        indices = list(shapes_layer.selected_data)
        default_value = shapes_layer.feature_defaults[label_feature][0]
        shapes_layer.features[label_feature][indices] = default_value
        shapes_layer.events.features()

    shapes_layer.events.feature_defaults.connect(set_selected_features_to_default)
    shapes_layer.events.features.connect(shapes_layer.refresh_text)

    def label_changed(value: str):
        """This is a callback that update the default features on the Shapes layer
        when the label menu selection changes
        """
        shapes_layer.feature_defaults[label_feature] = value
        shapes_layer.events.feature_defaults()

    label_menu.changed.connect(label_changed)

    return label_widget


# create a stack with the camera image shifted in each slice
n_slices = 5
base_image = data.camera()
image = np.zeros((n_slices, base_image.shape[0], base_image.shape[1]), dtype=base_image.dtype)
for slice_idx in range(n_slices):
    shift = 1 + 10 * slice_idx
    image[slice_idx, ...] = np.pad(base_image, ((0, 0), (shift, 0)), mode='constant')[:, :-shift]


# create a viewer with a fake t+2D image
viewer = napari.Viewer()
layer = viewer.add_image(image)

# create an empty shapes layer initialized with
# text set to display the box label
text_kwargs = {
    'string': text_feature,
    'size': text_size,
    'color': text_color
}
shapes = viewer.add_shapes(
    face_color='black',
    features=features,
    text=text_kwargs,
    ndim=3
)

# create the label section gui
label_widget = create_label_menu(
    shapes_layer=shapes,
    label_feature=text_feature,
    labels=box_annotations
)
# add the label selection gui to the viewer as a dock widget
viewer.window.add_dock_widget(label_widget, area='right', name='label_widget')

# set the shapes layer mode to adding rectangles
shapes.mode = 'add_rectangle'

if __name__ == '__main__':
    napari.run()



from __future__ import annotations

from typing import TYPE_CHECKING

from vispy.visuals.transforms import MatrixTransform, STTransform

from napari._vispy.utils.gl import BLENDING_MODES
from napari.utils.events import disconnect_events

if TYPE_CHECKING:
    from napari.layers import Layer
    from napari.utils.events import Event


class VispyBaseOverlay:
    """
    Base overlay backend for vispy.

    Creates event connections between napari Overlay models and the
    vispy backend, translating them into rendering.
    """

    def __init__(self, *, overlay, node, parent=None) -> None:
        super().__init__()
        self.overlay = overlay

        self.node = node
        self.node.order = self.overlay.order

        self.overlay.events.visible.connect(self._on_visible_change)
        self.overlay.events.opacity.connect(self._on_opacity_change)
        self.overlay.events.blending.connect(self._on_blending_change)

        if parent is not None:
            self.node.parent = parent

    def _on_visible_change(self) -> None:
        self.node.visible = self.overlay.visible

    def _on_opacity_change(self) -> None:
        self.node.opacity = self.overlay.opacity

    def _on_blending_change(self) -> None:
        self.node.set_gl_state(**BLENDING_MODES[self.overlay.blending])
        self.node.update()

    def reset(self) -> None:
        self._on_visible_change()
        self._on_opacity_change()
        self._on_blending_change()

    def close(self) -> None:
        disconnect_events(self.overlay.events, self)
        self.node.transforms = MatrixTransform()
        self.node.parent = None


class VispyCanvasOverlay(VispyBaseOverlay):
    """
    Vispy overlay backend for overlays that live in canvas space.

    NOTE: Subclasses must follow some rules:
    - ensure that when `_on_position_change` is called, the x_size and y_size
      attributes are already updated depending on the overlay size, to ensure
      proper tiling. Alternatively, override this method if the overlay is
      *not* supposed to be tiled
    - ensure that the napari Overlay model uses the `position` field correctly
      (must be a CanvasPosition enum if tileable, or anything else if "free")

    canvas_position_callback is set by the VispyCanvas object, and is responsible
    to update the position of all canvas overlays whenever necessary
    """

    def __init__(self, *, overlay, node, parent=None) -> None:
        super().__init__(overlay=overlay, node=node, parent=parent)
        self.x_size = 0.0
        self.y_size = 0.0
        self.node.transform = STTransform()
        self.overlay.events.position.connect(self._on_position_change)
        self.canvas_position_callback = lambda: None

    def _on_position_change(self, event: Event | None = None) -> None:
        # NOTE: when subclasses call this method, they should first ensure sizes
        # (x_size, and y_size) are set correctly
        self.canvas_position_callback()

    def reset(self) -> None:
        super().reset()
        self._on_position_change()

    def close(self) -> None:
        super().close()
        self.canvas_position_callback = lambda: None


class VispySceneOverlay(VispyBaseOverlay):
    """
    Vispy overlay backend for overlays that live in scene (2D or 3D) space.
    """

    def __init__(self, *, overlay, node, parent=None) -> None:
        super().__init__(overlay=overlay, node=node, parent=parent)
        self.node.transform = MatrixTransform()


class LayerOverlayMixin:
    def __init__(self, *, layer: Layer, overlay, node, parent=None) -> None:
        super().__init__(
            node=node,
            overlay=overlay,
            parent=parent,
        )
        self.layer = layer
        # need manual connection here because these overlays are not necessarily
        # always a child of the actual vispy node of the layer (eg, canvas overlays)
        self.layer.events.visible.connect(self._on_visible_change)

    def _on_visible_change(self) -> None:
        self.node.visible = self.overlay.visible and self.layer.visible

    def close(self) -> None:
        disconnect_events(self.layer.events, self)
        super().close()


class ViewerOverlayMixin:
    def __init__(self, *, viewer, overlay, node, parent=None) -> None:
        super().__init__(
            node=node,
            overlay=overlay,
            parent=parent,
        )
        self.viewer = viewer

    def close(self) -> None:
        disconnect_events(self.viewer.events, self)
        super().close()



from qtpy.QtWidgets import (
    QCheckBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr, checked_to_bool
from napari.layers import Tracks
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtHideCompletedTracksCheckBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the hide completed
    tracks attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    hide_completed_tracks_checkbox : qtpy.QtWidgets.QCheckBox
        Checkbox controlling if completed tracks should be hidden.
    hide_completed_tracks_checkbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for showing the option checkbox.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer

        # Setup widgets
        self.hide_completed_tracks_checkbox = QCheckBox()
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'hide_completed_tracks',
                self.hide_completed_tracks_checkbox,
                'setChecked',
            )
        )
        connect_setattr(
            self.hide_completed_tracks_checkbox.stateChanged,
            layer,
            'hide_completed_tracks',
            convert_fun=checked_to_bool,
        )

        self.hide_completed_tracks_checkbox_label = QtWrappedLabel(
            trans._('hide completed:')
        )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (
                self.hide_completed_tracks_checkbox_label,
                self.hide_completed_tracks_checkbox,
            )
        ]



from __future__ import annotations

from itertools import cycle
from random import sample
from typing import TYPE_CHECKING, Any

import numpy as np
from vispy.app.timer import Timer

from napari._vispy.overlays.base import ViewerOverlayMixin, VispyCanvasOverlay
from napari._vispy.visuals.welcome import Welcome
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.theme import get_theme

if TYPE_CHECKING:
    from vispy.scene import Node
    from vispy.util.event import Event

    from napari import Viewer
    from napari.components.overlays import Overlay


class VispyWelcomeOverlay(ViewerOverlayMixin, VispyCanvasOverlay):
    def __init__(
        self, *, viewer: Viewer, overlay: Overlay, parent: Node | None = None
    ) -> None:
        super().__init__(
            node=Welcome(), viewer=viewer, overlay=overlay, parent=parent
        )
        self.viewer.events.theme.connect(self._on_theme_change)
        self.viewer.layers.events.connect(self._on_visible_change)

        self.overlay.events.version.connect(self._on_version_change)
        self.overlay.events.shortcuts.connect(self._on_shortcuts_change)
        self.overlay.events.tips.connect(self._on_tips_change)

        self.node.canvas.native.resized.connect(self._on_position_change)

        self.tips_iterator = cycle(["You're awesome!"])
        self.tip_timer = Timer(10, self.next_tip)
        self.next_tip()

        self.reset()

    def _on_position_change(self, event: Any = None) -> None:
        if self.node.canvas is not None:
            x, y = np.array(self.node.canvas.size)
            self.node.set_scale_and_position(x, y)

    def _on_theme_change(self) -> None:
        if self.node.parent is not None and self.node.parent.canvas.bgcolor:
            background_color = self.node.parent.canvas.bgcolor.rgba
        else:
            background_color = get_theme(self.viewer.theme).canvas.as_hex()
            background_color = transform_color(background_color)[0]
        color = np.subtract(1, background_color)
        color[-1] = background_color[-1]
        color *= 0.8  # dim a bit
        self.node.set_color(color)

    def _on_visible_change(self) -> None:
        show = self.overlay.visible and not self.viewer.layers
        self.node.visible = show
        if show:
            self.tip_timer.start()
        else:
            self.tip_timer.stop()

    def _on_version_change(self) -> None:
        self.node.set_version(self.overlay.version)

    def _on_shortcuts_change(self) -> None:
        self.node.set_shortcuts(self.overlay.shortcuts)

    def _on_tips_change(self) -> None:
        if self.overlay.tips:
            self.tips_iterator = cycle(
                sample(self.overlay.tips, len(self.overlay.tips))
            )
        else:
            self.tips_iterator = cycle(["You're awesome!"])

    def next_tip(self, event: Event | None = None) -> None:
        self.node.set_tip(next(self.tips_iterator))

    def reset(self) -> None:
        super().reset()
        self._on_theme_change()
        self._on_version_change()
        self._on_shortcuts_change()
        self._on_tips_change()
        self.next_tip()

    def close(self) -> None:
        self.tip_timer.stop()
        super().close()



"""
Show points based on feature
============================

.. tags:: visualization-advanced
"""

import numpy as np
from magicgui import magicgui

import napari

# create points with a randomized "confidence" feature
points = np.random.rand(100, 3) * 100
colors = np.random.rand(100, 3)
confidence = np.random.rand(100)

viewer = napari.Viewer(ndisplay=3)
points = viewer.add_points(
        points, face_color=colors, features={'confidence': confidence}
        )


# create a simple widget with magicgui which provides a slider that controls the visibility
# of individual points based on their "confidence" value
@magicgui(
    auto_call=True,
    threshold={'widget_type': 'FloatSlider', 'min': 0, 'max': 1}
)
def confidence_slider(layer: napari.layers.Points, threshold=0.5):
    layer.shown = layer.features['confidence'] > threshold


viewer.window.add_dock_widget(confidence_slider)

if __name__ == '__main__':
    napari.run()



"""
Surface timeseries
==================

Display a surface timeseries using data from nilearn

.. tags:: experimental
"""
from importlib.metadata import version

from nilearn import datasets, surface
from packaging.version import parse

import napari

if parse(version("numpy")) >= parse('1.24') and parse(version("nilearn")) < parse('0.10.1'):
    raise RuntimeError(
        'Incompatible numpy version. '
        'You must have numpy less than 1.24 for nilearn 0.10.1 and below to '
        'work and download the example data'
    )


# Fetch datasets - this will download dataset if datasets are not found
nki_dataset = datasets.fetch_surf_nki_enhanced(n_subjects=1)
fsaverage = datasets.fetch_surf_fsaverage()

# Load surface data and resting state time series from nilearn
brain_vertices, brain_faces = surface.load_surf_data(fsaverage['pial_left'])
brain_vertex_depth = surface.load_surf_data(fsaverage['sulc_left'])
timeseries = surface.load_surf_data(nki_dataset['func_left'][0])
# nilearn provides data as n_vertices x n_timepoints, but napari requires the
# vertices axis to be placed last to match NumPy broadcasting rules
timeseries = timeseries.transpose((1, 0))

# create an empty viewer
viewer = napari.Viewer(ndisplay=3)

# add the mri
viewer.add_surface((brain_vertices, brain_faces, brain_vertex_depth), name='base')
viewer.add_surface((brain_vertices, brain_faces, timeseries),
                    colormap='turbo', opacity=0.9,
                    contrast_limits=[-1.5, 3.5], name='timeseries')

if __name__ == '__main__':
    napari.run()



"""
3D Paths
========

Display two vectors layers ontop of a 4-D image layer. One of the vectors
layers is 3D and "sliced" with a different set of vectors appearing on
different 3D slices. Another is 2D and "broadcast" with the same vectors
appearing on each slice.

.. tags:: visualization-advanced, layers
"""

import numpy as np
from skimage import data

import napari

blobs = data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=0.05
        )

viewer = napari.Viewer(ndisplay=3)
viewer.add_image(blobs.astype(float))

# sample vector coord-like data
path = np.array([np.array([[0, 0, 0], [0, 10, 10], [0, 5, 15], [20, 5, 15],
    [56, 70, 21], [127, 127, 127]]),
    np.array([[0, 0, 0], [0, 10, 10], [0, 5, 15], [0, 5, 15],
        [0, 70, 21], [0, 127, 127]])])

print('Path', path.shape)
layer = viewer.add_shapes(
    path, shape_type='path', edge_width=4, edge_color=['red', 'blue']
)

if __name__ == '__main__':
    napari.run()



import argparse
from timeit import default_timer

import numpy as np

import napari


def create_sample_coords(n_polys=3000, n_vertices=32):
    """random circular polygons with given number of vertices"""
    center = np.random.randint(0, 1000, (n_polys, 2))
    radius = (
        1000
        / np.sqrt(n_polys)
        * np.random.uniform(0.9, 1.1, (n_polys, n_vertices))
    )

    phi = np.linspace(0, 2 * np.pi, n_vertices, endpoint=False)
    rays = np.stack([np.sin(phi), np.cos(phi)], 1)

    radius = radius.reshape((-1, n_vertices, 1))
    rays = rays.reshape((1, -1, 2))
    center = center.reshape((-1, 1, 2))
    coords = center + radius * rays
    return coords


def time_me(label, func):
    # print(f'{name} start')
    t = default_timer()
    res = func()
    t = default_timer() - t
    print(f'{label}: {t:.4f} s')
    return res


parser = argparse.ArgumentParser(description='')

parser.add_argument(
    '-n',
    '--n_polys',
    type=int,
    default=5000,
    help='number of polygons to show',
)
parser.add_argument(
    '-t',
    '--type',
    type=str,
    default='path',
    choices=['path', 'path_concat', 'polygon', 'rectangle', 'ellipse'],
)
parser.add_argument(
    '-c',
    '--concat',
    action='store_true',
    help='concatenate all coordinates to a single mesh',
)
parser.add_argument(
    '-v', '--view', action='store_true', help='show napari viewer'
)
parser.add_argument(
    '--properties', action='store_true', help='add dummy shape properties'
)

args = parser.parse_args()

coords = create_sample_coords(args.n_polys)

if args.type == 'rectangle':
    coords = coords[:, [4, 20]]
elif args.type == 'ellipse':
    coords = coords[:, [0, 8, 16,22]]
elif args.type == 'path_concat':
    args.type = 'path'
    coords = coords.reshape((1, -1, 2))


print(f'number of polygons: {len(coords)}')
print(f'layer type: {args.type}')
print(f'properties: {args.properties}')

properties = {
    'class': (['A', 'B', 'C', 'D'] * (len(coords) // 4 + 1))[
        : len(coords)
    ],
}
color_cycle = ['blue', 'magenta', 'green']

kwargs = {
    'shape_type': args.type,
    'properties': properties if args.properties else None,
    'face_color': 'class' if args.properties else [1,1,1,1],
    'face_color_cycle': color_cycle,
    'edge_color': 'class' if args.properties else [1,1,1,1],
    'edge_color_cycle': color_cycle,
}

layer = time_me(
    'time to create layer',
    lambda: napari.layers.Shapes(coords, **kwargs),
)

if args.view:
    # add the image
    viewer = napari.Viewer()
    viewer.add_layer(layer)

if __name__ == '__main__':
    napari.run()



from typing import Literal

import numpy as np

__all__ = (
    'BoxArray',
    'CoordinateArray',
    'CoordinateArray2D',
    'CoordinateArray3D',
    'CoordinateDtype',
    'EdgeArray',
    'IndexArray',
    'IndexDtype',
    'ShapeColor',
    'ShapeColorArray',
    'ShapeColorDtype',
    'TriangleArray',
    'TriangleDtype',
    'ZOrderArray',
    'ZOrderDtype',
)


CoordinateDtype = np.float32
IndexDtype = np.int32
ShapeColorDtype = np.float32
TriangleDtype = np.int32
ZOrderDtype = np.int32

BoxArray = np.ndarray[tuple[Literal[9], int], np.dtype[np.float32]]
CoordinateArray2D = np.ndarray[
    tuple[int, Literal[2]], np.dtype[CoordinateDtype]
]
CoordinateArray3D = np.ndarray[
    tuple[int, Literal[3]], np.dtype[CoordinateDtype]
]
CoordinateArray = CoordinateArray2D | CoordinateArray3D
EdgeArray = np.ndarray[tuple[int, Literal[2]], np.dtype[np.int64]]
IndexArray = np.ndarray[tuple[int], np.dtype[IndexDtype]]
ShapeColor = np.ndarray[tuple[Literal[4]], np.dtype[ShapeColorDtype]]
ShapeColorArray = np.ndarray[tuple[int, Literal[4]], np.dtype[ShapeColorDtype]]
TriangleArray = np.ndarray[tuple[int, Literal[3]], np.dtype[TriangleDtype]]
ZOrderArray = np.ndarray[tuple[int], np.dtype[ZOrderDtype]]



"""
This script is called in pre-commit to ensure that certain extras in
pyproject.toml are kept in sync with dependency-groups.

It is to provide a transition period while we migrate from extras to
dependency-groups, and can be removed once the migration is complete.

It is planned to remote these extras and this script in the 0.7.0 release.
"""

import sys
from pathlib import Path

try:
    import tomllib
except ModuleNotFoundError:
    sys.exit('This script requires Python 3.11 or newer.')


SYNC_GROUPS = (
    'testing',
    'testing_extra',
    'gallery',
    'docs',
    'release',
    'dev',
    'build',
)


def include_group_to_extras(dependency: str | dict) -> str:
    if isinstance(dependency, str):
        return dependency
    return f'napari[{dependency["include-group"]}]'


def main():
    pyproject_toml_path = Path(__file__).parent.parent / 'pyproject.toml'
    data = tomllib.loads(pyproject_toml_path.read_text())
    extras = data['project']['optional-dependencies']
    groups = data['dependency-groups']

    for group_name in SYNC_GROUPS:
        group = {include_group_to_extras(x) for x in groups[group_name]}
        extra = set(extras[group_name])
        if group != extra:
            print(f"Mismatch in '{group_name}':")
            only_in_group = group - extra
            only_in_extra = extra - group
            if only_in_group:
                print(f'  In group but not in extras: {only_in_group}')
            if only_in_extra:
                print(f'  In extras but not in group: {only_in_extra}')
            return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""
Mixed dimensions labels
=======================

Overlay a 3D segmentation on a 4D time series.

Sometimes, our data have mixed dimensionality. napari "right-aligns" the
dimensions of your data, following NumPy broadcasting conventions [1]_. In this
example, we show how we can see a 3D segmentation overlaid on a 4D dataset. As
we slice through the dataset, the segmentation stays unchanged, but is visible
on every slice.

.. [1] https://numpy.org/doc/stable/user/basics.broadcasting.html

.. tags:: visualization-nD
"""

import numpy as np
from scipy import ndimage as ndi
from skimage.data import binary_blobs

import napari

blobs3d = binary_blobs(length=64, volume_fraction=0.1, n_dim=3).astype(float)

blobs3dt = np.stack([np.roll(blobs3d, 3 * i, axis=2) for i in range(10)])

labels = ndi.label(blobs3dt[5])[0]

viewer = napari.Viewer(ndisplay=3)

image_layer = viewer.add_image(blobs3dt)
labels_layer = viewer.add_labels(labels)
viewer.dims.current_step = (5, 0, 0, 0)

if __name__ == '__main__':
    napari.run()



"""
Image warping with mouse event callbacks
========================================

Warp an image based on points moved by user using mouse event callbacks in
napari.

Derived from scikit-image example: https://scikit-image.org/docs/stable/auto_examples/transform/plot_tps_deformation.html

This example is fully explained in the following tutorial from SciPy 2025:
https://napari.org/napari-scipy2025-workshop/notebooks/image_warping.html

Credit to Lars Grüter for the original idea and first attempts at implementation at SciPy 2024.

.. tags:: interactivity
"""

from functools import partial

import numpy as np
import skimage as ski

import napari

# Set up base image (to be warped) and points layers
image = ski.data.checkerboard()
src = np.array([[66, 66], [133, 66], [66, 133], [133, 133]])

viewer = napari.Viewer()
checkerboard_image_layer = viewer.add_image(image, name='checkerboard')
source_points_layer = viewer.add_points(
    src, name='source_points', symbol='+', face_color='red', size=5
)
moving_points_layer = viewer.add_points(src.copy(), name='moving_points')

# ensure moving_points layer is in Select mode
moving_points_layer.mode = 'select'


# Define a function to estimate the warping required to transform the destination points
# into the source points, and then apply the warping to the original image data,
# replacing the image layer data in-place.
def warp(
    im_layer: 'napari.layers.Image',
    original_image_data: 'np.ndarray',
    src: 'np.ndarray',
    dst: 'np.ndarray',
) -> None:
    # Warp image using thin-plate spline transformation from skimage.
    tps = ski.transform.ThinPlateSplineTransform()
    tps.estimate(dst, src)
    warped = ski.transform.warp(original_image_data, tps)
    # warped will be in 0..1 floats, so we need to
    # multiply by 255 to get it back to the same range
    # as the original data
    im_layer.data = (warped * 255).astype(original_image_data.dtype)


# Use partial to specify arguments for the warp function because the event
# callback only has access to the moving `points_layer` and the `event` object itself
warp_checkerboard = partial(
    warp,  # the warping function
    checkerboard_image_layer,  # im_layer argument
    image,  # original_image_data argument
    src,  # src argument
)


def warp_on_move(points_layer, event):
    # we do nothing here, as we don't care about the mouse press,
    # by yielding we have the chance to do things on drag
    yield

    # while the mouse is moving, we call our warp function
    while event.type == 'mouse_move':
        # ensure a point is selected and we're in select mode
        if not points_layer.selected_data or points_layer.mode != 'select':
            return
        # find the index into the points data of the currently selected point
        # we use the last selected point as that's likely what the mouse is hovering
        # over
        moved_point_index = list(points_layer.selected_data)[-1]
        # make a copy of the moving_points so original array is unchanged
        dst = points_layer.data.copy()
        # assign the current mouse position into the correct index to
        # update the location of the point
        dst[moved_point_index] = event.position
        # warp image
        warp_checkerboard(dst)
        yield

    # Empty here as nothing is done on mouse release


# Hook up callback to the layer event
moving_points_layer.mouse_drag_callbacks.append(warp_on_move)

if __name__ == '__main__':
    napari.run()



"""
Scale bar
=========

Display a 3D volume and the scale bar

.. tags:: experimental
"""
from skimage import data

import napari

cells = data.cells3d()

viewer = napari.Viewer(ndisplay=3)

viewer.add_image(
    cells,
    name=('membrane', 'nuclei'),
    channel_axis=1,
    scale=(0.29, 0.26, 0.26),
)
viewer.scale_bar.visible = True

# Text options
viewer.scale_bar.unit = 'um'  # set to None to diplay no unit
viewer.scale_bar.length = 23  # length, in units, of the scale bar
viewer.scale_bar.font_size = 20  # default is 10

# Text color
viewer.scale_bar.colored = True  # default value is False
viewer.scale_bar.color = 'yellow'  # default value is magenta: (1,0,1,1)

# Background box
viewer.scale_bar.box = True  # add background box, default is False
viewer.scale_bar.box_color = (0, 1, 1, 0.2)  # cyan with alpha=0.2

# Scale bar position
viewer.scale_bar.position = 'bottom_left'  # default is 'bottom_right'

if __name__ == '__main__':
    napari.run()



import numpy as np

from napari.layers import Tracks

from .utils import Skip


class TracksSuite:
    param_names = ['size', 'n_tracks']
    params = [(5 * np.power(10, np.arange(7))).tolist(), [1, 10, 100, 1000]]

    skip_params = Skip(
        if_in_pr=lambda size, n_tracks: size > 500 or n_tracks > 10,
        always=lambda size, n_tracks: n_tracks * 5 > size,
    )
    # we skip cases where the number of tracks times five is larger than the size as it is not useful

    def setup(self, size, n_tracks):
        """
        Create tracks data
        """
        rng = np.random.default_rng(0)

        track_ids = rng.integers(n_tracks, size=size)
        time = np.zeros(len(track_ids))

        for value, counts in zip(
            *np.unique(track_ids, return_counts=True), strict=False
        ):
            t = rng.permutation(counts)
            time[track_ids == value] = t

        coordinates = rng.uniform(size=(size, 3))

        data = np.concatenate(
            (track_ids[:, None], time[:, None], coordinates),
            axis=1,
        )

        self.data = data

        # create layer for the update benchmark
        self.layer = Tracks(self.data)

    def time_create_layer(self, *_) -> None:
        Tracks(self.data)

    def time_update_layer(self, *_) -> None:
        self.layer.data = self.data


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
Annotate 2D
===========

Display one points layer ontop of one image layer using the ``add_points`` and
``add_image`` APIs

.. tags:: analysis
"""

import numpy as np
from skimage import data

import napari

print('click to add points; close the window when finished.')

viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True)
points = viewer.add_points(np.zeros((0, 2)))
points.mode = 'add'

if __name__ == '__main__':
    napari.run()

    print('you clicked on:')
    print(points.data)



"""
nD points with features
=======================

Display one points layer ontop of one 4-D image layer using the
add_points and add_image APIs, where the markes are visible as nD objects
across the dimensions, specified by their size

.. tags:: visualization-nD, features-table
"""

import numpy as np
from skimage import data

import napari

blobs = data.binary_blobs(
    length=100, blob_size_fraction=0.05, n_dim=3, volume_fraction=0.05
)
viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float))

# create the points
points = []
for z in range(blobs.shape[0]):
    points += [[z, 25, 25], [z, 25, 75], [z, 75, 25], [z, 75, 75]]

# create the features for setting the face and edge color.
face_feature = np.array(
    [True, True, True, True, False, False, False, False]
    * int(blobs.shape[0] / 2)
)
border_feature = np.array(['A', 'B', 'C', 'D', 'E'] * int(len(points) / 5))

features = {
    'face_feature': face_feature,
    'border_feature': border_feature,
}

points_layer = viewer.add_points(
    points,
    features=features,
    size=3,
    border_width=5,
    border_width_is_relative=False,
    border_color='border_feature',
    face_color='face_feature',
    out_of_slice_display=False,
)

# change the face color cycle
points_layer.face_color_cycle = ['white', 'black']

# change the border_color cycle.
# there are 4 colors for 5 categories, so 'c' will be recycled
points_layer.border_color_cycle = ['c', 'm', 'y', 'k']

if __name__ == '__main__':
    napari.run()



from math import ceil

from napari.utils.events import EventedModel


def empty(event):
    pass


def fibonacci(n):
    if n < 2:
        return n
    return fibonacci(n - 1) + fibonacci(n - 2)


class Model(EventedModel):
    a: int = 3
    b: float = 2.0
    c: int = 3

    @property
    def d(self):
        return (self.c + self.a) ** self.b

    @d.setter
    def d(self, value):
        self.c = value
        self.a = value
        self.b = value * 1.1

    @property
    def e(self):
        return (fibonacci(self.c) + fibonacci(self.a)) ** fibonacci(
            ceil(self.b)
        )


class EventedModelSuite:
    """Benchmarks for EventedModel."""

    def setup(self):
        self.model = Model()
        self.model.events.a.connect(empty)
        self.model.events.b.connect(empty)
        self.model.events.c.connect(empty)
        self.model.events.e.connect(empty)

    def time_event_firing(self):
        self.model.d = 4
        self.model.d = 18

    def time_long_connection(self):
        def long_connection(event):
            for _i in range(5):
                fibonacci(event.source.c)

        self.model.events.e.connect(long_connection)

        self.model.d = 15


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



from __future__ import annotations

import argparse
import logging
import os
import re
import subprocess  # nosec
import sys
from pathlib import Path

from tomllib import loads

REPO_DIR = Path(__file__).parent.parent
DEFAULT_NAME = 'auto-dependency-upgrades'


def get_base_branch_name(ref_name, event):
    if ref_name == DEFAULT_NAME:
        return 'main'
    if ref_name.startswith(DEFAULT_NAME):
        if event in {'pull_request', 'pull_request_target'}:
            return os.environ.get('GITHUB_BASE_REF')
        return ref_name[len(DEFAULT_NAME) + 1 :]
    return ref_name


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--main-packages', action='store_true')
    args = parser.parse_args()

    ref_name = get_ref_name()
    event = os.environ.get('GITHUB_EVENT_NAME', '')

    base_branch = get_base_branch_name(ref_name, event)

    try:
        res = get_changed_dependencies(base_branch, not args.main_packages)
    except ValueError as e:
        print(e)
        sys.exit(1)

    if args.main_packages:
        print('\n'.join(f' * {x}' for x in sorted(res)))
    elif res:
        print(', '.join(f'`{x}`' for x in res))
    else:
        print('only indirect updates')


def get_branches() -> list[str]:
    """
    Get all branches from the repository.
    """
    out = subprocess.run(  # nosec
        ['git', 'branch', '--list', '--format', '%(refname:short)', '-a'],
        capture_output=True,
        check=True,
    )
    return out.stdout.decode().split('\n')


def calc_changed_packages(
    base_branch: str, src_dir: Path, python_version: str
) -> list[str]:
    """
    Calculate a list of changed packages based on python_version

    Parameters
    ----------
    base_branch: str
        branch against which to compare
    src_dir: Path
        path to the root of the repository
    python_version: str
        python version to use

    Returns
    -------
    list[str]
        list of changed packages
    """
    changed_name_re = re.compile(r'\+([\w-]+)')

    command = [
        'git',
        'diff',
        base_branch,
        str(
            src_dir
            / 'resources'
            / 'constraints'
            / f'constraints_py{python_version}.txt'
        ),
    ]
    logging.info('Git diff call: %s', ' '.join(command))
    try:
        out = subprocess.run(  # nosec
            command,
            capture_output=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise ValueError(
            f'git diff failed with return code {e.returncode}'
            ' stderr: {e.stderr.decode()!r}'
            ' stdout: {e.stdout.decode()!r}'
        ) from e

    return [
        changed_name_re.match(x)[1].lower()
        for x in out.stdout.decode().split('\n')
        if changed_name_re.match(x)
    ]


def get_ref_name() -> str:
    """
    Get the name of the current branch.
    """
    ref_name = os.environ.get('GITHUB_REF_NAME')
    if ref_name:
        return ref_name
    out = subprocess.run(  # nosec
        ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
        capture_output=True,
        check=True,
    )
    return out.stdout.decode().strip()


def calc_only_direct_updates(
    changed_packages: list[str], src_dir: Path
) -> list[str]:
    name_re = re.compile(r'[\w-]+')

    metadata = loads((src_dir / 'pyproject.toml').read_text())['project']
    optional_dependencies = metadata['optional-dependencies']

    packages = (
        metadata['dependencies']
        + optional_dependencies['pyqt5']
        + optional_dependencies['pyqt6']
        + optional_dependencies['pyside6']
        + optional_dependencies['testing']
        + optional_dependencies['all']
    )
    packages = [
        name_re.match(package).group().lower()
        for package in packages
        if name_re.match(package)
    ]
    return sorted(set(packages) & set(changed_packages))


def get_changed_dependencies(
    base_branch: str,
    all_packages=False,
    python_version='3.10',
    src_dir: Path | None = None,
):
    """
    Get the changed dependencies.

    all_packages: bool
        If True, return all packages, not just the direct dependencies.
    """
    if src_dir is None:
        src_dir = Path(__file__).parent.parent

    branches = get_branches()

    if base_branch not in branches:
        if f'origin/{base_branch}' not in branches:
            raise ValueError(
                f'base branch {base_branch} not found in {branches!r}'
            )
        base_branch = f'origin/{base_branch}'

    changed_packages = calc_changed_packages(
        base_branch, src_dir, python_version=python_version
    )

    if all_packages:
        return sorted(set(changed_packages))

    return calc_only_direct_updates(changed_packages, src_dir)


if __name__ == '__main__':
    main()



# /// script
# dependencies = [
#   "imageio",
#   "tifffile",
# ]
# ///

import re
import sys
from pathlib import Path
from pprint import pformat

import imageio
import tifffile

BUILTINS_YAML_PATH = (
    Path(__file__).parent.parent / 'src' / 'napari_builtins' / 'builtins.yaml'
)

re_exts = re.compile(
    r'(filename_patterns:\s+\[\n)([^]]*)(\n\s+\])', flags=re.DOTALL
)


def gen_extensions():
    imageio_exts = {
        f'*{ext}' for fmt in imageio.formats for ext in fmt.extensions
    }
    imageio_video_exts = {
        f'*{fmt.extension}'
        for fmt in imageio.config.video_extensions
        if any(name == 'FFMPEG' for name in fmt.priority)
    }
    tifffile_exts = {f'*.{ext}' for ext in tifffile.TIFF.FILE_EXTENSIONS}
    napari_exts = {f'*.{ext}' for ext in ('npy', 'zarr', 'csv')}

    exts = sorted(
        imageio_exts | imageio_video_exts | tifffile_exts | napari_exts
    )

    json = pformat(exts, compact=True, indent=10)[:-1].replace('[', ' ')
    return json


def get_extensions():
    with open(BUILTINS_YAML_PATH) as f:
        contents = f.read()

    match = re.search(re_exts, contents)
    return match.group(2)


def replace_extensions(new):
    with open(BUILTINS_YAML_PATH) as f:
        contents = f.read()

    with open(BUILTINS_YAML_PATH, 'w') as f:
        f.write(re.sub(re_exts, rf'\1{new}\3', contents))


if __name__ == '__main__':
    print('Checking reader extensions...')

    CI = '--ci' in sys.argv
    old = get_extensions()
    new = gen_extensions()
    if old != new:
        print('Extension patterns are not up to date!')
        if CI:
            print('Updating extensions...')
            replace_extensions(new)
            print('Extensions updated.')
            sys.exit(0)
        else:
            print('Here is the new extension block:\n')
            print(new)
            sys.exit(1)

    print('Extension patterns are up to date.')



"""Example of using low-level QtNodeTreeView with Node and Group

:class:`napari.utils.tree.Node` is a class that may be used as a mixin that
allows an object to be a member of a "tree".

:class:`napari.utils.tree.Group` is a (nestable) mutable sequence of Nodes, and
is also itself a Node (this is the "composite" pattern):
https://refactoring.guru/design-patterns/composite/python/example

These two classes may be used to create tree-like data structures that behave
like pure python lists of lists.

This examples shows that :class:`napari._qt.containers.QtNodeTreeView`
is capable of providing a basic GUI for any tree structure based on
`napari.utils.tree.Group`.
"""
import napari
from napari._qt.containers import QtNodeTreeView
from napari.qt import get_qapp
from napari.utils.tree import Group, Node

get_qapp()

# create a group of nodes.
root = Group(
    [
        Node(name='6'),
        Group(
            [
                Node(name='1'),
                Group([Node(name='2'), Node(name='3')], name='g2'),
                Node(name='4'),
                Node(name='5'),
                Node(name='tip'),
            ],
            name='g1',
        ),
        Node(name='7'),
        Node(name='8'),
        Node(name='9'),
    ],
    name='root',
)
# create Qt view onto the Group
view = QtNodeTreeView(root)
# show the view
view.show()


# pretty __str__ makes nested tree structure more interpretable
print(root)
# root
#   ├──6
#   ├──g1
#   │  ├──1
#   │  ├──g2
#   │  │  ├──2
#   │  │  └──3
#   │  ├──4
#   │  ├──5
#   │  └──tip
#   ├──7
#   ├──8
#   └──9


# spy on events
root.events.reordered.connect(lambda e: print('reordered to: ', e.value))
root.selection.events.changed.connect(
    lambda e: print(
        f'selection changed.  added: {e.added}, removed: {e.removed}'
    )
)
root.selection.events._current.connect(
    lambda e: print(f'current item changed to: {e.value}')
)

if __name__ == '__main__':
    napari.run()



import numpy as np

from napari._qt.layer_controls.qt_surface_controls import QtSurfaceControls
from napari.layers import Surface
from napari.layers.surface._surface_constants import SHADING_TRANSLATION

data = np.array([[0, 0], [0, 20], [10, 0], [10, 10]])
faces = np.array([[0, 1, 2], [1, 2, 3]])
values = np.linspace(0, 1, len(data))
_SURFACE = (data, faces, values)


def test_shading_combobox(qtbot):
    layer = Surface(_SURFACE)
    qtctrl = QtSurfaceControls(layer)
    qtbot.addWidget(qtctrl)
    assert (
        qtctrl._shading_combobox_control.shading_combobox.currentText()
        == layer.shading
    )

    for display, shading in SHADING_TRANSLATION.items():
        qtctrl._shading_combobox_control.shading_combobox.setCurrentText(
            display
        )
        assert layer.shading == shading

    for display, shading in SHADING_TRANSLATION.items():
        layer.shading = shading
        assert (
            qtctrl._shading_combobox_control.shading_combobox.currentText()
            == display
        )



import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
)
from superqt import QLabeledSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Labels
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtBrushSizeSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current brush
    size attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    brush_size_slider : superqt.QLabeledDoubleSlider
        Slider controlling current brush size of the layer.
    brush_size_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the brush size chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.brush_size.connect(self._on_brush_size_change)

        # Setup widgets
        sld = QLabeledSlider(Qt.Orientation.Horizontal)
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        sld.setMinimum(1)
        sld.setMaximum(40)
        sld.setSingleStep(1)
        connect_setattr(sld.valueChanged, self._layer, 'brush_size')
        self.brush_size_slider = sld
        self._on_brush_size_change()

        self.brush_size_slider_label = QtWrappedLabel(trans._('brush size:'))

    def _on_brush_size_change(self) -> None:
        """Receive layer model brush size change event and update the slider."""
        with qt_signals_blocked(self.brush_size_slider):
            value = self._layer.brush_size
            value = np.maximum(1, int(value))
            if value > self.brush_size_slider.maximum():
                self.brush_size_slider.setMaximum(int(value))
            self.brush_size_slider.setValue(value)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.brush_size_slider_label, self.brush_size_slider)]



"""
Clipping planes interactive
===========================

Display a 3D image (plus labels) with a clipping plane and interactive controls
for moving the plane

.. tags:: experimental
"""
import numpy as np
from scipy import ndimage
from skimage import data
from vispy.geometry import create_sphere

import napari

viewer = napari.Viewer(ndisplay=3)

# VOLUME and LABELS
blobs = data.binary_blobs(
    length=64, volume_fraction=0.1, n_dim=3
).astype(float)

labeled = ndimage.label(blobs)[0]

plane_parameters = {
    'position': (32, 32, 32),
    'normal': (1, 1, 1),
    'enabled': True
}

volume_layer = viewer.add_image(
    blobs, rendering='mip', name='volume',
    experimental_clipping_planes=[plane_parameters],
)

labels_layer = viewer.add_labels(
    labeled, name='labels', blending='translucent',
    experimental_clipping_planes=[plane_parameters],
)

# POINTS
points_layer = viewer.add_points(
    np.random.rand(20, 3) * 64, size=5,
    experimental_clipping_planes=[plane_parameters],
)

# SPHERE
mesh = create_sphere(method='ico')
sphere_vert = mesh.get_vertices() * 20
sphere_vert += 32
surface_layer = viewer.add_surface(
    (sphere_vert, mesh.get_faces()),
    experimental_clipping_planes=[plane_parameters],
)

# SHAPES
shapes_data = np.random.rand(3, 4, 3) * 64

shapes_layer = viewer.add_shapes(
    shapes_data,
    face_color=['magenta', 'green', 'blue'],
    experimental_clipping_planes=[plane_parameters],
)

# VECTORS
vectors = np.zeros((20, 2, 3))
vectors[:, 0] = 32
vectors[:, 1] = (np.random.rand(20, 3) - 0.5) * 32

vectors_layer = viewer.add_vectors(
    vectors,
    experimental_clipping_planes=[plane_parameters],
)


def point_in_bounding_box(point, bounding_box):
    return bool(np.all(point > bounding_box[0]) and np.all(point < bounding_box[1]))


@viewer.mouse_drag_callbacks.append
def shift_plane_along_normal(viewer, event):
    """Shift a plane along its normal vector on mouse drag.
    This callback will shift a plane along its normal vector when the plane is
    clicked and dragged. The general strategy is to
    1) find both the plane normal vector and the mouse drag vector in canvas
    coordinates
    2) calculate how far to move the plane in canvas coordinates, this is done
    by projecting the mouse drag vector onto the (normalised) plane normal
    vector
    3) transform this drag distance (canvas coordinates) into data coordinates
    4) update the plane position
    It will also add a point to the points layer for a 'click-not-drag' event.
    """
    # get layers from viewer
    volume_layer = viewer.layers['volume']

    # Calculate intersection of click with data bounding box
    near_point, _far_point = volume_layer.get_ray_intersections(
        event.position,
        event.view_direction,
        event.dims_displayed,
    )

    # Calculate intersection of click with plane through data
    intersection = volume_layer.experimental_clipping_planes[0].intersect_with_line(
        line_position=near_point, line_direction=event.view_direction
    )

    # Check if click was on plane by checking if intersection occurs within
    # data bounding box. If so, exit early.
    if not point_in_bounding_box(intersection, volume_layer.extent.data):
        return

    # Get plane parameters in vispy coordinates (zyx -> xyz)
    plane_normal_data_vispy = np.array(volume_layer.experimental_clipping_planes[0].normal)[[2, 1, 0]]
    plane_position_data_vispy = np.array(volume_layer.experimental_clipping_planes[0].position)[[2, 1, 0]]

    # Get transform which maps from data (vispy) to canvas
    # note that we're using a private attribute here, which may not be present in future napari versions
    visual2canvas = viewer.window._qt_viewer.canvas.layer_to_visual[volume_layer].node.get_transform(
        map_from='visual', map_to='canvas'
    )

    # Find start and end positions of plane normal in canvas coordinates
    plane_normal_start_canvas = visual2canvas.map(plane_position_data_vispy)
    plane_normal_end_canvas = visual2canvas.map(plane_position_data_vispy + plane_normal_data_vispy)

    # Calculate plane normal vector in canvas coordinates
    plane_normal_canv = (plane_normal_end_canvas - plane_normal_start_canvas)[[0, 1]]
    plane_normal_canv_normalised = (
            plane_normal_canv / np.linalg.norm(plane_normal_canv)
    )

    # Disable interactivity during plane drag
    volume_layer.mouse_pan = False
    labels_layer.mouse_pan = False
    labels_layer.mouse_pan = False
    points_layer.mouse_pan = False
    surface_layer.mouse_pan = False
    shapes_layer.mouse_pan = False
    vectors_layer.mouse_pan = False

    # Store original plane position and start position in canvas coordinates
    original_plane_position = volume_layer.experimental_clipping_planes[0].position
    start_position_canv = event.pos

    yield
    while event.type == 'mouse_move':
        # Get end position in canvas coordinates
        end_position_canv = event.pos

        # Calculate drag vector in canvas coordinates
        drag_vector_canv = end_position_canv - start_position_canv

        # Project the drag vector onto the plane normal vector
        # (in canvas coordinates)
        drag_projection_on_plane_normal = np.dot(
            drag_vector_canv, plane_normal_canv_normalised
        )

        # Update position of plane according to drag vector
        # only update if plane position is within data bounding box
        drag_distance_data = drag_projection_on_plane_normal / np.linalg.norm(plane_normal_canv)
        updated_position = original_plane_position + drag_distance_data * np.array(
            volume_layer.experimental_clipping_planes[0].normal)

        if point_in_bounding_box(updated_position, volume_layer.extent.data):
            volume_layer.experimental_clipping_planes[0].position = updated_position
            labels_layer.experimental_clipping_planes[0].position = updated_position
            points_layer.experimental_clipping_planes[0].position = updated_position
            surface_layer.experimental_clipping_planes[0].position = updated_position
            shapes_layer.experimental_clipping_planes[0].position = updated_position
            vectors_layer.experimental_clipping_planes[0].position = updated_position

        yield

    # Re-enable
    volume_layer.mouse_pan = True
    labels_layer.mouse_pan = True
    points_layer.mouse_pan = True
    surface_layer.mouse_pan = True
    shapes_layer.mouse_pan = True
    vectors_layer.mouse_pan = True


viewer.axes.visible = True
viewer.camera.angles = (45, 45, 45)
viewer.camera.zoom = 5
viewer.text_overlay.update({
    'text': 'Drag the clipping plane surface to move it along its normal.',
    'font_size': 20,
    'visible': True,
})

if __name__ == '__main__':
    napari.run()



from warnings import warn

from napari.utils.translations import trans

try:
    from qtpy import API_NAME, QtCore
except Exception as e:
    if 'No Qt bindings could be found' in str(e):
        from importlib.metadata import version
        from inspect import cleandoc

        from napari.utils._env_detection import detect_environment

        raise ImportError(
            trans._(
                cleandoc(
                    """
                No Qt bindings could be found for napari=={version}.

                napari requires either PyQt5 (default), PyQt6 or PySide6 to be installed in the environment.

                With pip, you can install either with:
                  $ pip install -U 'napari[all]'  # default choice
                  $ pip install -U 'napari[pyqt5]'
                  $ pip install -U 'napari[pyqt6]'
                  $ pip install -U 'napari[pyside6]'

                With conda, you need to do:
                  $ conda install -c conda-forge pyqt
                  $ conda install -c conda-forge pyside6

                Our heuristics suggest you are using '{tool}' to manage your packages.
                """
                ),
                deferred=True,
                tool=detect_environment().value,
                version=version('napari'),
            )
        ) from e
    raise


# When QT is not the specific version, we raise a warning:
if tuple(int(x) for x in QtCore.__version__.split('.')[:3]) < (5, 12, 3):
    import importlib.metadata

    try:
        dist_info_version = importlib.metadata.version(API_NAME)
        if dist_info_version != QtCore.__version__:
            warn_message = trans._(
                "\n\nIMPORTANT:\nYou are using QT version {version}, but version {dversion} was also found in your environment.\nThis usually happens when you 'conda install' something that also depends on PyQt\n*after* you have pip installed napari (such as jupyter notebook).\nYou will likely run into problems and should create a fresh environment.\nIf you want to install conda packages into the same environment as napari,\nplease add conda-forge to your channels: https://conda-forge.org\n",
                deferred=True,
                version=QtCore.__version__,
                dversion=dist_info_version,
            )
    except ModuleNotFoundError:
        warn_message = trans._(
            '\n\nnapari was tested with QT library `>=5.12.3`.\nThe version installed is {version}. Please report any issues with\nthis specific QT version at https://github.com/Napari/napari/issues.',
            deferred=True,
            version=QtCore.__version__,
        )
    warn(message=warn_message, stacklevel=1)


from napari._qt.qt_event_loop import get_qapp, quit_app, run
from napari._qt.qt_main_window import Window

__all__ = ['Window', 'get_qapp', 'quit_app', 'run']



"""
Add multiscale image
====================

Displays a multiscale image

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data
from skimage.transform import pyramid_gaussian

import napari

# create multiscale from astronaut image
base = np.tile(data.astronaut(), (8, 8, 1))
multiscale = list(
    pyramid_gaussian(base, downscale=2, max_layer=4, channel_axis=-1)
)
print('multiscale level shapes: ', [p.shape[:2] for p in multiscale])

# add image multiscale
viewer = napari.Viewer()
layer = viewer.add_image(multiscale, multiscale=True)

if __name__ == '__main__':
    napari.run()



import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QPainter
from qtpy.QtWidgets import (
    QHBoxLayout,
    QWidget,
)
from superqt import QLargeIntSpinBox

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr, qt_signals_blocked
from napari.layers import Labels
from napari.layers.labels._labels_utils import get_dtype
from napari.utils._dtype import get_dtype_limits
from napari.utils.events import disconnect_events
from napari.utils.translations import trans


class QtColorBox(QWidget):
    """A widget that shows a square with the current label color.

    Parameters
    ----------
    layer : napari.layers.Labels
        An instance of a napari Labels layer.
    """

    def __init__(self, layer: Labels) -> None:
        super().__init__()

        self._layer = layer
        self._layer.events.selected_label.connect(
            self._on_selected_label_change
        )
        self._layer.events.opacity.connect(self._on_opacity_change)
        self._layer.events.colormap.connect(self._on_colormap_change)

        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        self._height = 24
        self.setFixedWidth(self._height)
        self.setFixedHeight(self._height)
        self.setToolTip(trans._('Selected label color'))

        self.color = None

    def _on_selected_label_change(self) -> None:
        """Receive layer model label selection change event & update colorbox."""
        self.update()

    def _on_opacity_change(self) -> None:
        """Receive layer model label selection change event & update colorbox."""
        self.update()

    def _on_colormap_change(self) -> None:
        """Receive label colormap change event & update colorbox."""
        self.update()

    def paintEvent(self, event) -> None:
        """Paint the colorbox.  If no color, display a checkerboard pattern.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        painter = QPainter(self)
        if self._layer._selected_color is None:
            self.color = None
            for i in range(self._height // 4):
                for j in range(self._height // 4):
                    if (i % 2 == 0 and j % 2 == 0) or (
                        i % 2 == 1 and j % 2 == 1
                    ):
                        painter.setPen(QColor(230, 230, 230))
                        painter.setBrush(QColor(230, 230, 230))
                    else:
                        painter.setPen(QColor(25, 25, 25))
                        painter.setBrush(QColor(25, 25, 25))
                    painter.drawRect(i * 4, j * 4, 5, 5)
        else:
            color = np.round(255 * self._layer._selected_color).astype(int)
            painter.setPen(QColor(*list(color)))
            painter.setBrush(QColor(*list(color)))
            painter.drawRect(0, 0, self._height, self._height)
            self.color = tuple(color)

    def disconnect_widget_controls(self) -> None:
        disconnect_events(self._layer.events, self)

    def deleteLater(self) -> None:
        self.disconnect_widget_controls()
        super().deleteLater()

    def closeEvent(self, event) -> None:
        """Disconnect events when widget is closing."""
        self.disconnect_widget_controls()
        super().closeEvent(event)


class QtLabelControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current label
    layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    colorbox : QtColorBox
        Widget that shows current layer label color.
    label_color : qtpy.QtWidget.QWidget
        Wrapper widget for the selection_spinbox and colorbox widgets.
    label_color_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the label chooser widget.
    selection_spinbox : superqt.QLargeIntSpinBox
        Widget to select a specific label by its index.
        N.B. cannot represent labels > 2**53.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.data.connect(self._on_data_change)

        # Setup widgets
        self.selection_spinbox = QLargeIntSpinBox()
        dtype_lims = get_dtype_limits(get_dtype(layer))
        self.selection_spinbox.setRange(*dtype_lims)
        self.selection_spinbox.setValue(self._layer.selected_label)
        self.selection_spinbox.setKeyboardTracking(False)
        self.selection_spinbox.valueChanged.connect(self.change_selection)
        self.selection_spinbox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'selected_label',
                self.selection_spinbox,
                'setValue',
            )
        )

        self.label_color_label = QtWrappedLabel(trans._('label:'))
        self.label_color = QWidget()
        self.label_color.setProperty('emphasized', True)
        color_layout = QHBoxLayout()
        color_layout.setContentsMargins(0, 2, 0, 1)
        self.colorbox = QtColorBox(layer)
        color_layout.addWidget(self.colorbox)
        color_layout.addWidget(self.selection_spinbox)
        self.label_color.setLayout(color_layout)

    def change_selection(self, value: int) -> None:
        """Change currently selected label.
        Parameters
        ----------
        value : int
            Index of label to select.
        """
        self._layer.selected_label = value
        self.selection_spinbox.clearFocus()
        # TODO: decouple
        self.parent().setFocus()

    def _on_data_change(self) -> None:
        """Update label selection spinbox min/max when data changes."""
        with qt_signals_blocked(self.selection_spinbox):
            dtype_lims = get_dtype_limits(get_dtype(self._layer))
            self.selection_spinbox.setRange(*dtype_lims)

    def disconnect_widget_controls(self) -> None:
        self.colorbox.disconnect_widget_controls()
        super().disconnect_widget_controls()

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.label_color_label, self.label_color)]
